import keyring
import requests
import typer

from mirageml.constants import SERVICE_ID, NOTION_SYNC_ENDPOINT

def sync_plugin(args):
    plugin_name = args["plugin"]
    if plugin_name == "notion":
        notion_provider_token = keyring.get_password(SERVICE_ID, "notion_provider_token")
        if not notion_provider_token:
            typer.secho("Please add the notion plugin first. Run `mirageml add plugin notion`", fg=typer.colors.BRIGHT_RED, bold=True)
            return
        else:
            access_token = keyring.get_password(SERVICE_ID, "access_token")
            headers = {
                "Authorization": f"Bearer {access_token}",
            }
            sync_response = requests.post(NOTION_SYNC_ENDPOINT, json={}, headers=headers)
            sync_response_data = sync_response.json()
            if "error" in sync_response_data:
                typer.secho(sync_response_data["error"], fg=typer.colors.BRIGHT_RED, bold=True)
            else:
                typer.secho("Syncing notion triggered successfully. You will receive an email when the sync is complete.", fg=typer.colors.BRIGHT_GREEN, bold=True)
    else:
        typer.secho("Only notion plugin is supported for now.", fg=typer.colors.BRIGHT_RED, bold=True)
        return




def add_indices_to_code_blocks(markdown):
    """Add indices with enhanced formatting before each code block."""
    import re
    code_blocks = re.findall(r'```(.*?)```', markdown, re.DOTALL)
    for index, block in enumerate(code_blocks, start=1):
        markdown = markdown.replace(f"```{block}```", f"\n_**Block {index}**_:\n```{block}```", 1)
    return markdown

def extract_code_from_markdown(markdown):
    """Extract code blocks from markdown without the language specifier."""
    import re
    # Capture everything between triple backticks
    code_blocks = re.findall(r'```(.*?)```', markdown, re.DOTALL)

    # Remove potential language specifiers (e.g., "python") from the start of each block
    cleaned_code_blocks = []
    for block in code_blocks:
        cleaned_block = "\n".join(block.split("\n")[1:]) if block.split("\n")[0].isalpha() else block
        cleaned_code_blocks.append(cleaned_block.strip())

    return cleaned_code_blocks

def copy_code_to_clipboard(code_blocks, selected_indices):
    import typer
    import pyperclip

    selected_code = '\n\n'.join([code_blocks[i-1] for i in selected_indices if 1 <= i <= len(code_blocks)])

    if selected_code:
        pyperclip.copy(selected_code)
        typer.secho("Selected code blocks copied to clipboard!", fg=typer.colors.BRIGHT_GREEN, bold=True)



from mirageml.classes import LoginManager
from mirageml.constants import REDIRECT_URI

def login():
  m = LoginManager(
     provider="google",
     provider_options={
      "redirect_to": REDIRECT_URI,
      'query_params': {
        'access_type': 'offline',
        'prompt': 'consent',
      }
    }
  )
  m.start_web_server()
  m.open_browser()



import os
import json
import typer

def show_config():
    config = load_config()
    from rich import print
    print(json.dumps(config, indent=4))

def load_config():
    config_path = os.path.expanduser("~/.mirageml.json")
    if os.path.exists(config_path):
        with open(config_path, 'r') as f:
            return json.load(f)
    return {
        "local_mode": False,
        "model": "gpt-4"
    }

def save_config(config):
    config_path = os.path.expanduser("~/.mirageml.json")
    with open(config_path, 'w') as f:
        json.dump(config, f, indent=4)

def set_config():
    config = load_config()

    valid = {
        "local_mode": (("True", "False"), json.loads),
        "model": (("gpt-3.5-turbo", "gpt-4"), str)
    }

    for key, value in config.items():
        curvalue = value
        while True:
            question = f"Enter the value for '{key}' [{', '.join(valid[key][0])}] (current value: {curvalue}): "
            value = input(question)
            if value == "": break
            if value not in valid[key][0]:
                typer.secho(f"Invalid value for '{key}'. Please use one of these options {valid[key][0]}", fg=typer.colors.BRIGHT_RED)
                continue
            config[key] = valid[key][1](value.lower())
            save_config(config)
            break

    typer.secho("MirageML Config updated!", fg=typer.colors.BRIGHT_GREEN)
    show_config()

def set_var_config(json_data):
    config = load_config()
    for key, value in json_data.items():
        config[key] = value
    save_config(config)



import os
from .utils.vectordb import exists_qdrant_db, create_qdrant_db
from .utils.web_source import crawl_website
from .utils.local_source import crawl_files
from .config import load_config


def add_web_source(name, link, remote=False):
    config = load_config()
    remote = False if config["local_mode"] else True
    print(f"Indexing {link}...")
    # if exists_qdrant_db(collection_name=name):
    #     print(f"Collection already exists, please delete with `mirageml delete source {name}` first.")
    #     return

    data, metadata = crawl_website(link)
    create_qdrant_db(data, metadata, collection_name=name, remote=remote)

def add_local_source(name=None, path=None):
    config = load_config()
    remote = False if config["local_mode"] else True
    print("Indexing Local Files...")
    # Crawl the files under the current directory
    data, metadata = crawl_files()
    # collection_name should be absolute path
    collection_name = os.path.abspath('.').replace('/', '_') if name is None else name
    collection_name = collection_name.replace(' ', '_')
    create_qdrant_db(data, metadata, collection_name=collection_name, remote=remote)
    return collection_name

def add_source(name, link):
    name = name.replace(' ', '_')
    if link:
        add_web_source(name, link)
    else:
        add_local_source()



import typer

def set_sources():
    from .utils.vectordb import list_qdrant_db, list_remote_qdrant_db
    from .config import set_var_config
    set_var_config({
        "local": list_qdrant_db(),
        "remote": list_remote_qdrant_db()
    })

def list_sources():
    from .utils.vectordb import list_qdrant_db, list_remote_qdrant_db
    from .config import set_var_config

    sources = list_qdrant_db()

    if len(sources) != 0:
        typer.secho("Local Sources:", fg=typer.colors.BRIGHT_GREEN, bold=True)
        print("* " + "\n* ".join(sources + ["local (this will index the files in your current directory)"]))
    remote_sources = list_remote_qdrant_db()
    print("------------------")
    if len(remote_sources) != 0:
        typer.secho("Remote Sources:", fg=typer.colors.BRIGHT_GREEN, bold=True)
        print("* " + "\n* ".join(remote_sources))

    set_var_config({
        "local": sources,
        "remote": remote_sources
    })



import os

def crawl_files(start_dir="."):
    file_data = []

    # Walk through the directory structure
    for dirpath, dirnames, filenames in os.walk(os.path.abspath(start_dir)):
        for filename in filenames:
            # Skip hidden files
            if filename.startswith('.') or dirpath.split('/')[-1].startswith('.'):
                continue
            filepath = os.path.join(dirpath, filename)

            try:
                # Read the file content
                with open(filepath, 'r', encoding='utf-8') as file:
                    file_content = file.read()
                    file_data.append((file_content, filepath))
            except Exception as e:
                # If unable to read a file, you can print an error or continue to the next file
                pass

    data = [x[1] + ": " + x[0] for x in file_data]
    metadata = [dict({"data": x[0]}, **{"source": x[1]}) for x in file_data]
    return data, metadata



import os
import sys
import requests
import keyring

from io import StringIO
from ...constants import get_headers, SERVICE_ID, VECTORDB_EMBED_ENDPOINT, LLM_GPT_ENDPOINT

PACKAGE_DIR = os.path.dirname(__file__)
os.environ['TRANSFORMERS_CACHE'] = os.path.join(PACKAGE_DIR, 'models')

def local_get_embedding(text_list, embedding_model_id="BAAI/bge-small-en-v1.5"):
    from sentence_transformers import SentenceTransformer
    model_dir = os.path.join(PACKAGE_DIR, 'models', embedding_model_id)
    if not os.path.exists(model_dir):
        os.makedirs(model_dir, exist_ok=True)
        print("Downloading model to:", model_dir)
        print("This will take a few minutes and only happen once!")

    # Suppress stdout/stderr
    original_stdout, original_stderr = sys.stdout, sys.stderr
    sys.stdout, sys.stderr = StringIO(), StringIO()

    model = SentenceTransformer(
        embedding_model_id,
        cache_folder=model_dir
        )
    embeddings = model.encode(text_list, normalize_embeddings=False)

    # Restore stdout/stderr
    sys.stdout, sys.stderr = original_stdout, original_stderr

    # Convert the embeddings to a list
    embeddings = embeddings.tolist()
    return embeddings

def local_llm_call(messages, llm_model_id="TheBloke/Llama-2-7b-Chat-GGUF", stream=False):
    from ctransformers import AutoModelForCausalLM
    model_dir = os.path.join(PACKAGE_DIR, 'models', llm_model_id)
    if not os.path.exists(model_dir):
        os.makedirs(model_dir, exist_ok=True)
        print("Downloading model to:", model_dir)
        print("This will take a few minutes and only happen once!")

    # Suppress stdout/stderr
    original_stdout, original_stderr = sys.stdout, sys.stderr
    sys.stdout, sys.stderr = StringIO(), StringIO()

    llm = AutoModelForCausalLM.from_pretrained(
        "TheBloke/Llama-2-7b-Chat-GGUF",
        model_file="llama-2-7b-chat.Q4_K_M.gguf",
        model_type="llama",
    )

    # Restore stdout/stderr
    sys.stdout, sys.stderr = original_stdout, original_stderr
    formatted_messages = "\n".join([x["content"] for x in messages])

    if stream: return llm(formatted_messages, stream=True)
    else: return llm(formatted_messages)

def get_embedding(text_list, model="BAAI/bge-small-en-v1.5", local=False):
    if local: return local_get_embedding(text_list, embedding_model_id=model)
    response = requests.post(VECTORDB_EMBED_ENDPOINT, json={'data': text_list}, headers=get_headers())
    response.raise_for_status()  # Raise an exception if the request failed
    return response.json()['embedding']

def llm_call(messages, model="gpt-3.5-turbo", stream=False, local=False):
    if local: return local_llm_call(messages, stream=stream)
    json_data = {
        "model": model,
        "messages": messages,
        "stream": stream
    }
    return requests.post(LLM_GPT_ENDPOINT, json=json_data, headers=get_headers(), stream=stream)



from .utils.vectordb import delete_qdrant_db, delete_remote_qdrant_db

def delete_source(name: str):
    import typer
    from .config import load_config
    config = load_config()
    sources = config["local"] + config["remote"]

    while True:
        if name not in sources:
            typer.secho(f"Source: {name} does not exist. Choose one of the following or exit:", fg=typer.colors.BRIGHT_RED)
            print("\n".join(sources))
            name = typer.prompt("Enter source name", default="exit", show_default=False)
            if not name or name == "exit":
                return
        else: break

    typer.secho(f"Deleting Source: {name}...", fg=typer.colors.BRIGHT_RED)
    delete_qdrant_db(name)
    delete_remote_qdrant_db(name)
    print(f"Deleted Source: {name}")


from prompt_toolkit import prompt
from prompt_toolkit.keys import Keys
from prompt_toolkit.key_binding import KeyBindings

# Create custom key bindings
kb = KeyBindings()

# You can put your custom finish conditon function here
@kb.add(Keys.Escape, Keys.Enter)
def _(event):
    buffer = event.current_buffer
    buffer.insert_text('\n')

@kb.add(Keys.Enter)
def _(event):
    event.current_buffer.validate_and_handle()

def multiline_input(message):
    import typer
    typer.secho(f'{message} (Use Option+Enter for newline):', fg=typer.colors.BRIGHT_GREEN, bold=True)
    user_input = prompt("> ", multiline=True, key_bindings=kb)
    return user_input

if __name__ == '__main__':
    user_input = multiline_input("test_file")
    print(user_input)


import os
import re
import json
import time
import uuid
import requests
import tiktoken
import threading
from tqdm import tqdm
from qdrant_client import QdrantClient
from qdrant_client.http.models import PointStruct, VectorParams, Distance

import keyring
from ..list_sources import set_sources
from ...constants import get_headers, SERVICE_ID, VECTORDB_SEARCH_ENDPOINT, VECTORDB_LIST_ENDPOINT, VECTORDB_CREATE_ENDPOINT, VECTORDB_DELETE_ENDPOINT

from ..config import load_config


import typer
from rich.progress import Progress

PACKAGE_DIR = os.path.dirname(__file__)

progress = Progress()

def get_qdrant_db():
    QDRANT_LOCKFILE_PATH = os.path.join(PACKAGE_DIR, ".lock")
    if os.path.exists(QDRANT_LOCKFILE_PATH): os.remove(QDRANT_LOCKFILE_PATH)
    return QdrantClient(path=PACKAGE_DIR)

def exists_qdrant_db(collection_name="test"):
    qdrant_client = get_qdrant_db()
    return collection_name in [x.name for x in qdrant_client.get_collections().collections]

def create_remote_qdrant_db(data, metadata, collection_name="test"):
    json_data = {
        "user_id": keyring.get_password(SERVICE_ID, 'user_id'),
        "data": data,
        "metadata": metadata,
        "collection_name": collection_name
    }

    from rich.console import Console
    from rich.live import Live
    from rich.panel import Panel

    console = Console()
    with Live(Panel("Creating Embeddings...", title="[bold green]Indexer[/bold green]", border_style="green"),
                    console=console, transient=True, auto_refresh=True, vertical_overflow="visible") as live:

        response = requests.post(VECTORDB_CREATE_ENDPOINT, json=json_data, headers=get_headers(), stream=True)
        if response.status_code == 200:
            for chunk in response.iter_content(chunk_size=None):
                # process line here
                live.update(Panel(f"Indexing: {chunk.decode()}", title="[bold green]Indexer[/bold green]", border_style="green"))

    typer.secho(f"Created Source: {collection_name}", fg=typer.colors.GREEN, bold=True)
    set_sources()
    return True

def create_qdrant_db(data, metadata, collection_name="test", remote=False):
    from .brain import get_embedding

    if remote: return create_remote_qdrant_db(data, metadata, collection_name=collection_name)
    config = load_config()
    qdrant_client = get_qdrant_db()

    final_data, final_metadata = [], []

    from rich.console import Console
    from rich.live import Live
    from rich.panel import Panel

    console = Console()
    with Live(Panel("Indexing files...", title="[bold green]Indexer[/bold green]", border_style="green"),
                    console=console, transient=True, auto_refresh=True, vertical_overflow="visible") as live:

        # For each data chunk it based on number of tokens
        enc = tiktoken.get_encoding("cl100k_base")
        for curr_data, curr_metadata in zip(data, metadata):
            split_data = re.split(r"(?<=\.)\s+", curr_data)
            i, chunks, meta = 0, [], []
            while i < len(split_data):
                curr_chunk = ""
                while i < len(split_data) and len(enc.encode(str(curr_chunk))) < 1000:
                    curr_chunk += " " + split_data[i]
                    i += 1
                chunks.append(curr_chunk)
                meta.append({"data": curr_chunk, "source": curr_metadata["source"]})
            final_data.extend(chunks)
            final_metadata.extend(meta)

        live.update(Panel(f"Creating Embeddings...", title="[bold green]Indexer[/bold green]", border_style="green"))

        vectors = get_embedding(final_data, local=config["local_mode"])
        size = 384 if config["local_mode"] else 1536

        qdrant_client.recreate_collection(
            collection_name=collection_name,
            vectors_config=VectorParams(size=size, distance=Distance.COSINE),
        )

        for vector, f_metadata in zip(vectors, final_metadata):
            filepath = f_metadata["source"]

            live.update(Panel(f"Indexing: {filepath}", title="[bold green]Indexer[/bold green]", border_style="green"))

            qdrant_client.upsert(
                collection_name=collection_name,
                points=[
                    PointStruct(vector=vector, payload=f_metadata, id=uuid.uuid4().hex)
                ]
            )

    typer.secho(f"Created Source: {collection_name}", fg=typer.colors.GREEN, bold=True)

    set_sources()
    return qdrant_client

def list_remote_qdrant_db():
    json_data = {
        "user_id": keyring.get_password(SERVICE_ID, 'user_id'),
    }
    response = requests.post(VECTORDB_LIST_ENDPOINT, json=json_data, headers=get_headers())
    response.raise_for_status()  # Raise an exception if the request failed
    return response.json()

def list_qdrant_db():
    QDRANT_JSON_PATH = os.path.join(PACKAGE_DIR, "meta.json")
    if os.path.exists(QDRANT_JSON_PATH):
        with open(QDRANT_JSON_PATH) as json_file:
            collection = json.load(json_file)
            collection_names = list(collection["collections"].keys())
    else:
        collection_names = []
    return collection_names

def remote_qdrant_search(source_name, user_input):
    json_data = {
        "user_id": keyring.get_password(SERVICE_ID, 'user_id'),
        "collection_name": source_name,
        "search_query": user_input
    }
    response = requests.post(VECTORDB_SEARCH_ENDPOINT, json=json_data, headers=get_headers())
    response.raise_for_status()  # Raise an exception if the request failed
    set_sources()
    return response.json()

def qdrant_search(source_name, user_input):
    from .brain import get_embedding

    config = load_config()
    qdrant_client = get_qdrant_db()

    hits = qdrant_client.search(
                limit=5,
                collection_name=source_name,
                query_vector=get_embedding([user_input], local=config["local_mode"])[0],
            )
    hits = [{
        "score": hit.score,
        "payload": hit.payload
    } for hit in hits]
    return hits


def delete_remote_qdrant_db(collection_name="test"):
    json_data = {
        "user_id": keyring.get_password(SERVICE_ID, 'user_id'),
        "collection_name": collection_name,
    }
    response = requests.post(VECTORDB_DELETE_ENDPOINT, json=json_data, headers=get_headers())
    response.raise_for_status()  # Raise an exception if the request failed
    set_sources()
    return response.json()

def delete_qdrant_db(collection_name="test"):
    qdrant_client = get_qdrant_db()
    qdrant_client.delete_collection(collection_name=collection_name)
    set_sources()


import re
import typer
from rich.box import HORIZONTALS
from rich.markdown import Markdown
from rich.console import Console
from rich.live import Live
from rich.panel import Panel
from rich.prompt import Prompt

from .config import load_config
from .add_source import add_local_source
from .utils.brain import llm_call
from .utils.multiline import multiline_input
from .utils.codeblocks import extract_code_from_markdown, add_indices_to_code_blocks, copy_code_to_clipboard
from .utils.vectordb import list_qdrant_db, list_remote_qdrant_db, qdrant_search, remote_qdrant_search
from .utils.prompt_templates import RAG_TEMPLATE

console = Console()
config = load_config()

def search(live, user_input, sources): # ,local_sources, remote_sources=None):
    hits = []
    local = list_qdrant_db()
    remote = list_remote_qdrant_db()

    local_sources = [source for source in sources if source in local]
    remote_sources = [source for source in sources if source in remote]

    for source_name in local_sources:
        live.update(Panel("Searching through local sources...", title="[bold blue]Assistant[/bold blue]", border_style="blue"))
        try:
            # Search for matches in each source based on user input
            hits.extend(qdrant_search(source_name, user_input))
        except:
            # Handle potential errors with mismatched embedding models
            error_msg_local = f"Source: {source_name} was created with OpenAI's embedding model. Please run with `local_mode=False` or reindex with `mirageml delete source {source_name}; mirageml add source {source_name}`."
            error_msg_openai = f"Source: {source_name} was created with a local embedding model. Please run with `local_mode=True` or reindex with `mirageml delete source {source_name}; mirageml add source {source_name}`."

            typer.secho(error_msg_local if config["local_mode"] else error_msg_openai, fg=typer.colors.RED, bold=True)
            return

    for source_name in remote_sources:
        live.update(Panel("Searching through remote sources...", title="[bold blue]Assistant[/bold blue]", border_style="blue"))
        hits.extend(remote_qdrant_search(source_name, user_input))

    return hits

def rank_hits(hits):
    # Rank the hits based on their relevance
    sorted_hits = sorted(hits, key=lambda x: x["score"], reverse=True)[:5]
    return sorted_hits

def create_context(sorted_hits):
    return "\n\n".join([str(x["payload"]["source"]) + ": " + x["payload"]["data"] for x in sorted_hits])

def search_and_rank(live, user_input, sources):
    hits = search(live, user_input, sources)
    sorted_hits = rank_hits(hits)
    return sorted_hits


## MAIN FUNCTION
def rag_chat(file_or_url, sources):
    # Beginning of the chat sequence
    if file_or_url:
        from .utils.web_source import extract_file_or_url
        file_url_source, file_or_url_data = extract_file_or_url(file_or_url)

    if "local" in sources:
        sources.remove("local")
        sources.append(add_local_source())

    while True:
        try:
            user_input = multiline_input(f"Ask a question over these sources ({', '.join(sources)})")
            if user_input.lower().strip() == 'exit':
                typer.secho("Ending chat. Goodbye!", fg=typer.colors.BRIGHT_GREEN, bold=True)
                return
        except KeyboardInterrupt:
            typer.secho("Ending chat. Goodbye!", fg=typer.colors.BRIGHT_GREEN, bold=True)
            return

        # Live display while searching for relevant sources
        with Live(Panel("Searching through the relevant sources...",
                    title="[bold blue]Assistant[/bold blue]", border_style="blue", box=HORIZONTALS),
                    console=console, screen=False, auto_refresh=True, vertical_overflow="visible") as live:

            sorted_hits = search_and_rank(live, user_input, sources)
            context = create_context(sorted_hits)

            if file_or_url:
                context += "\n\n" + file_url_source + ": " + file_or_url_data
                sources.append(file_url_source)

            # Chat history that will be sent to the AI model
            chat_history = [
                {"role": "system", "content": "You are a helpful assistant that responds to questions concisely with the given context in the following format:\n{answer}\n\nSources:\n{sources}"},
                {"role": "user", "content": RAG_TEMPLATE.format(context=context, question=user_input, sources=sources)}
            ]

            # Fetch the AI's response
            ai_response = ""
            live.update(Panel("Found relevant sources! Answering question...", title="[bold blue]Assistant[/bold blue]", box=HORIZONTALS, border_style="blue"))
            response = llm_call(chat_history, model=config["model"], stream=True, local=config["local_mode"])

            if config["local_mode"]:
                for chunk in response:
                    ai_response += chunk
                    live.update(Panel(Markdown(ai_response), title="[bold blue]Assistant[/bold blue]", box=HORIZONTALS, border_style="blue"))
            else:
                for chunk in response.iter_content(1024):
                    if chunk:
                        decoded_chunk = chunk.decode('utf-8')
                        ai_response += decoded_chunk
                        live.update(Panel(Markdown(ai_response), title="[bold blue]Assistant[/bold blue]", box=HORIZONTALS, border_style="blue"))
            chat_history.append({"role": "assistant", "content": ai_response})

        while True:
            # Loop for follow-up questions
            try:
                code_blocks = extract_code_from_markdown(ai_response)
                if code_blocks:
                    selected_indices = multiline_input("Enter the numbers of the code blocks you want to copy, separated by commas (e.g. '1,3,4') or ask a follow-up. Type reset to search again")
                    try:
                        selected_indices = list(map(int, selected_indices.split(',')))
                        copy_code_to_clipboard(code_blocks, selected_indices)
                        ai_response = ""
                        continue
                    except ValueError: user_input = selected_indices
                else:
                    user_input = multiline_input("Ask a follow-up. Type reset to search again")
                if user_input.lower().strip() == 'exit':
                    typer.secho("Ending chat. Goodbye!", fg=typer.colors.BRIGHT_GREEN, bold=True)
                    return
                elif user_input.lower().strip() == 'reset':
                    break
            except KeyboardInterrupt:
                typer.secho("Ending chat. Goodbye!", fg=typer.colors.BRIGHT_GREEN, bold=True)
                return

            chat_history.append({"role": "user", "content": user_input})

            with Live(Panel("Assistant is typing...", title="[bold blue]Assistant[/bold blue]", box=HORIZONTALS, border_style="blue"),
                    console=console, transient=True, auto_refresh=True, refresh_per_second=8) as live:

                response = llm_call(chat_history, model=config["model"], stream=True, local=config["local_mode"])

                ai_response = ""
                if config["local_mode"]:
                    for chunk in response:
                        ai_response += chunk
                        live.update(Panel(Markdown(ai_response), title="[bold blue]Assistant[/bold blue]", box=HORIZONTALS, border_style="blue"))
                else:
                    for chunk in response.iter_content():
                        if chunk:
                            decoded_chunk = chunk.decode('utf-8')
                            ai_response += decoded_chunk
                            live.update(Panel(Markdown(ai_response), title="[bold blue]Assistant[/bold blue]", box=HORIZONTALS, border_style="blue"))

            chat_history.append({"role": "assistant", "content": ai_response})
            indexed_ai_response = add_indices_to_code_blocks(ai_response)
            console.print(Panel(Markdown(indexed_ai_response), title="[bold blue]Assistant[/bold blue]", box=HORIZONTALS, border_style="blue"))



"""
This file is the entry point for the commands module.
It initializes the module and imports the necessary modules.
"""

from .login import login
from .config import set_config, show_config

from .chat import normal_chat
from .rag import rag_chat

from .list_plugins import list_plugins
from .list_sources import list_sources

from .add_plugin import add_plugin
from .add_source import add_source

from .delete_source import delete_source

from .sync_plugin import sync_plugin


import typer

from mirageml.constants import REDIRECT_URI
from mirageml.classes import LoginManager

def add_gdrive():
    m = LoginManager(
        handler="google_auth_handler",
        provider="google",
        provider_options={
            "redirect_to": REDIRECT_URI,
            "scopes": "https://www.googleapis.com/auth/drive",
            "query_params": {
                "access_type": "offline",
                "prompt": "consent",
            }
        }
    )
    m.start_web_server()
    m.open_browser()

def add_notion():
    m = LoginManager(
        handler="notion_auth_handler",
        provider="notion",
        provider_options={
            "redirect_to": REDIRECT_URI,
        }
    )
    m.start_web_server()
    m.open_browser()

def add_plugin(args):
    plugin_name = args["plugin"]
    if plugin_name == "notion":
        add_notion()
    # elif plugin_name == "gdrive":
    #     add_gdrive()
    else:
        typer.secho("Only notion plugin is supported for now.", fg=typer.colors.BRIGHT_RED, bold=True)


import requests
from bs4 import BeautifulSoup
import urllib.parse
from urllib.parse import urlparse, urlunparse
import os
import tempfile
import subprocess
import shutil
import os
from langchain.document_loaders import AsyncChromiumLoader
from langchain.document_transformers import BeautifulSoupTransformer

import typer
from rich.markdown import Markdown
from rich.console import Console
from rich.live import Live
from rich.panel import Panel
from rich.table import Table
from rich.tree import Tree
from rich.progress import track

import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.CRITICAL)
logging.basicConfig(level=logging.CRITICAL)
logging.disable(logging.CRITICAL)
logging.getLogger("langchain").setLevel(logging.CRITICAL)

console = Console()

def tree_to_paths(node, path=None):
    path = path or []
    for child in node.children:
        yield from tree_to_paths(child, path + [child.label])
    if path:
        yield path

def crawl_with_playwright(live, start_url):
    urls = {start_url: set()}
    visited_links = set()
    from playwright.sync_api import sync_playwright

    def get_all_play_links(page):
        # Execute script on the page to get all links
        links = page.eval_on_selector_all("a", '''(anchors) => {
            return anchors.map(anchor => anchor.href);
        }''')
        for link in links:
            parsed_link = urlparse(link)
            cleaned_link = urlunparse((parsed_link.scheme, parsed_link.netloc, parsed_link.path, "", "", ""))
            yield cleaned_link

    with sync_playwright() as p:
        browser = p.chromium.launch()
        page = browser.new_page()
        page.goto(start_url)
        to_visit = [start_url]

        while to_visit:
            current_url = to_visit.pop(0)
            if current_url not in visited_links:
                visited_links.add(current_url)
                live.update(Panel(f"Scraping: {current_url}", title="[bold green]Scraper[/bold green]", border_style="green"))
                page.goto(current_url)
                links_on_page = get_all_play_links(page)
                for link in links_on_page:
                    if link.startswith(start_url):
                        to_visit.append(link)
        browser.close()
    return visited_links

def get_all_links(live, url):
    response = requests.get(url, headers={'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.3'})
    soup = BeautifulSoup(response.text, 'html.parser')
    a_tags = soup.find_all('a')
    if a_tags:
        for link in soup.find_all('a'):
            absolute_link = urllib.parse.urljoin(url, link.get('href'))
            parsed_link = urlparse(absolute_link)
            cleaned_link = urlunparse((parsed_link.scheme, parsed_link.netloc, parsed_link.path, "", "", ""))
            if cleaned_link.startswith(url):  # Only yield child URLs
                live.update(Panel(f"Scraping: {cleaned_link}", title="[bold green]Scraper[/bold green]", border_style="green"))
                yield cleaned_link
    else:
        for link in crawl_with_playwright(live, url):
            yield link

def check_playwright_chromium():
    try:
        # Try to launch chromium and close immediately. This is just to check its presence.
        subprocess.run(['python3', '-m', 'playwright', 'chromium', '--version'], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        return True
    except (subprocess.CalledProcessError, FileNotFoundError):
        print("Playwright Chromium required for webscraping.")
        print("Please run `python3 -m playwright install chromium`")
        return False

def crawl_website(start_url):
    if not start_url.endswith("/"): start_url += "/"
    urls = {start_url: set()}
    visited_links = set()
    to_visit = [start_url]

    # Check if playwright is installed
    if not check_playwright_chromium():
        return

    while to_visit:
        with Live(Panel("Preparing to Scrape...", title="[bold green]Scraper[/bold green]", border_style="green"),
                    console=console, transient=True, auto_refresh=True, vertical_overflow="visible") as live:
            with tempfile.TemporaryDirectory() as tmp_dir:
                # print(f"Created temporary directory: {tmp_dir}")
                while to_visit:
                    current_url = to_visit.pop(0)
                    if current_url not in visited_links:
                        visited_links.add(current_url)
                        for link in get_all_links(live, current_url):
                            to_visit.append(link)
        # pretty print all of the subpages that were indexed and ask the user if they want to continue
        typer.secho("Subpaths Per URL:", fg=typer.colors.GREEN, bold=True)
        print()

        # Process each link and add to tree
        for url in visited_links:
            # Check which key in urls the visited link is under
            url_key = ""
            for key in urls.keys():
                if key in url: url_key = key

            if not url_key: continue

            try:
                url = url.replace(url_key, "").split("/")[0]
                if url: urls[url_key].add(url)
            except: continue

        # Print the results
        for domain, paths in urls.items():
            typer.secho(f"{domain}:", fg=typer.colors.BRIGHT_GREEN, bold=True)
            if len (paths) == 0:
                print("No subpaths found, you may have to manually add the url when prompted.")
                print()
                continue
            for path in sorted(paths):
                print(path)
            print()


        user_input = input("Do you want to index another URL? Enter a URL or leave empty (default: no): ")
        user_input = user_input.strip()
        if user_input and not user_input.lower().startswith('n'):
            if user_input.lower().startswith('https://'):
                link = user_input
                if not link.endswith("/"): link += "/"
                to_visit = [link]
                urls[link] = set()
            else:
                while True:
                    link = input("Link for the source (exit to skip): ")
                    if link.lower().strip() == 'exit': break
                    elif not link.startswith("https://"):
                        typer.echo("Please enter a valid link starting with https://")
                        continue
                    break
            if link.lower().strip() != 'exit':
                if not link.endswith("/"): link += "/"
                to_visit = [link]
                urls[link] = set()
            else: to_visit = []
        else:
            to_visit = []

    data, metadata = [], []
    with Live(Panel("Preparing to Clean", title="[bold green]Cleaner[/bold green]", border_style="green"),
                    console=console, transient=True, auto_refresh=True, vertical_overflow="visible") as live:
        for link in visited_links:
            loader = AsyncChromiumLoader([link])
            live.update(Panel(f"Cleaning: {link}", title="[bold green]Cleaner[/bold green]", border_style="green"))
            html = loader.load()
            bs_transformer = BeautifulSoupTransformer()
            docs_transformed = bs_transformer.transform_documents(html, tags_to_extract=["p", "li", "div", "a", "code"])
            data.extend([x.page_content for x in docs_transformed])
            metadata.extend([dict({"data": x.page_content}, **x.metadata) for x in docs_transformed])
    return data, metadata

def extract_file_or_url(file_or_url):
    if not file_or_url.startswith("http"):
        try:
            # Read the file content
            with open(file_or_url, 'r', encoding='utf-8') as file:
                file_content = file.read()
                source, file_or_url_data = file_or_url, file_or_url + ": " + file_content
                typer.secho(f"Loaded file: {file_or_url}", fg=typer.colors.BRIGHT_GREEN, bold=True)
        except Exception as e:
            typer.secho(f"Unable to read file: {file_or_url}", fg=typer.colors.BRIGHT_RED, bold=True)
    else:
        try:
            if not check_playwright_chromium(): return
            from langchain.document_loaders import AsyncChromiumLoader
            from langchain.document_transformers import BeautifulSoupTransformer
            loader = AsyncChromiumLoader([file_or_url])
            html = loader.load()
            bs_transformer = BeautifulSoupTransformer()
            docs_transformed = bs_transformer.transform_documents(html, tags_to_extract=["p", "li", "div", "a", "code"])
            source, file_or_url_data = docs_transformed[0].metadata["source"], docs_transformed[0].page_content
            typer.secho(f"Loaded webpage: {file_or_url}", fg=typer.colors.BRIGHT_GREEN, bold=True)
        except Exception as e:
            typer.secho(f"Unable to read url make sure that the url starts with http: {file_or_url}", fg=typer.colors.BRIGHT_RED, bold=True)

    return source, file_or_url_data


if __name__ == "__main__":
    import time
    start_time = time.time()
    crawl_website("https://segment.com/docs/")
    print(f"Time taken: {time.time() - start_time}")


from supabase import create_client

SERVICE_ID = "mirageml"

PORT = 9998
REDIRECT_URI = f"http://localhost:{PORT}/callback"
SUPABASE_URL = "https://kfskvbhwrwpbruczecka.supabase.co"
SUPABASE_KEY = "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpc3MiOiJzdXBhYmFzZSIsInJlZiI6Imtmc2t2Ymh3cndwYnJ1Y3plY2thIiwicm9sZSI6ImFub24iLCJpYXQiOjE2OTcwNTI5MDYsImV4cCI6MjAxMjYyODkwNn0.KZoU1QMfvh72QfKJK_HD3EQ_246g-QbggSzPbT83Su8"
NOTION_SYNC_ENDPOINT = " https://mirageml--notion-sync-trigger-notion-sync.modal.run"
ANALYTICS_WRITE_KEY = "WeKYF6EZAYtVxUwlx8g3sG1JKDMMv6jY"

VECTORDB_EMBED_ENDPOINT = "https://mirageml--vectordb-embed-text.modal.run"
LLM_GPT_ENDPOINT = "https://mirageml--llm-gpt.modal.run"

VECTORDB_SEARCH_ENDPOINT = 'https://mirageml--vectordb-search-db.modal.run'
VECTORDB_LIST_ENDPOINT = "https://mirageml--vectordb-list-db.modal.run"
VECTORDB_CREATE_ENDPOINT = "https://mirageml--vectordb-create-db.modal.run"
VECTORDB_DELETE_ENDPOINT = "https://mirageml--vectordb-delete-db.modal.run"

supabase = create_client(SUPABASE_URL, SUPABASE_KEY)

def fetch_new_access_token():
    import keyring
    refresh_token = keyring.get_password(SERVICE_ID, 'refresh_token')
    response = supabase.auth._refresh_access_token(refresh_token)
    session = response.session
    keyring.set_password(SERVICE_ID, 'access_token', session.access_token)
    keyring.set_password(SERVICE_ID, 'refresh_token', session.refresh_token)
    keyring.set_password(SERVICE_ID, 'expires_at', str(session.expires_at))
    return session.access_token

def get_headers():
    import time
    import typer
    import keyring
    expires_at = keyring.get_password(SERVICE_ID, 'expires_at')
    access_token = keyring.get_password(SERVICE_ID, 'access_token')
    if expires_at and float(expires_at) < time.time():
        try:
            refresh_token = keyring.get_password(SERVICE_ID, 'refresh_token')
            response = supabase.auth._refresh_access_token(refresh_token)
            session = response.session
            keyring.set_password(SERVICE_ID, 'access_token', session.access_token)
            keyring.set_password(SERVICE_ID, 'refresh_token', session.refresh_token)
            keyring.set_password(SERVICE_ID, 'expires_at', str(session.expires_at))
            access_token = session.access_token
        except Exception as e:
            print(e)
            typer.echo("Please login again. Run `mirageml login`")
            raise typer.Exit()
    headers = {
        "Authorization": f"Bearer {access_token}"
    }
    return headers

def help_list_sources():
    import os
    import sys
    import json
    invoked_alias = sys.argv[0].split('/')[-1]  # Extract only the alias name

    config_path = os.path.expanduser("~/.mirageml.json")
    config = {}
    if os.path.exists(config_path):
        with open(config_path, 'r') as f:
            config = json.load(f)

    local_sources = config.get("local", [])
    remote_sources = config.get("remote", [])
    all_sources = list(set(local_sources + remote_sources))
    if len(all_sources) == 0:
        final_string = "Specify sources to use as context:\n\n\n**"+ invoked_alias +" chat -s {source1} -s {source2}**\n\n\n\n"
    elif len(all_sources) == 1:
        final_string = f"Specify sources to use as context:\n\n\nEx: **{invoked_alias} chat -s {all_sources[0]}**\n\n\n\n"
    else:
        final_string = f"Specify sources to use as context:\n\n\nEx: **{invoked_alias} chat -s {all_sources[0]} -s {all_sources[1]}**\n\n\n\n"

    local_sources.append("local (this will index the files in your current directory)")
    if len(local_sources) != 0:
        final_string += "**Local Sources:**\n\n"
        final_string +="\n\n* ".join(local_sources)
    final_string +="\n\n---\n\n"
    if len(remote_sources) != 0:
        final_string += "**Remote Sources:**\n\n"
        final_string +="\n\n* ".join(remote_sources)
    return final_string



RAG_TEMPLATE = """Answer the question based on the following context. If the context is relevant, mention which sources you used:
{context}

Sources:
{sources}

Question: {question}
"""

CHAT_TEMPLATE = """Answer the question using the following context:
{context}

Question: {question}
"""


import sys
import time
import typer
from typing import List

app = typer.Typer(
    no_args_is_help=True,
    add_completion=False,
    rich_markup_mode="markdown",
    help="""
    MirageML CLI

    **You can use 'mirage' or 'mml' to call the package.**



    See the website at https://mirageml.com/ for documentation and more information
    about running code on MirageML.
    """,
)

config_app = typer.Typer(name="config", help="Manage the config", no_args_is_help=True)
add_app = typer.Typer(name="add", help="Add a new resource", no_args_is_help=True)
list_app = typer.Typer(name="list", help="List resources", no_args_is_help=True)
sync_app = typer.Typer(name="sync", help="Sync resources", no_args_is_help=True)
delete_app = typer.Typer(name="delete", help="Delete resources", no_args_is_help=True)

app.add_typer(config_app, rich_help_panel="Utils and Configs")
app.add_typer(add_app, rich_help_panel="Manage Resources")
app.add_typer(list_app, rich_help_panel="Manage Resources")
app.add_typer(sync_app, rich_help_panel="Manage Resources")
app.add_typer(delete_app, rich_help_panel="Manage Resources")


@app.callback()
def main(ctx: typer.Context):
    import keyring
    from .constants import SERVICE_ID, ANALYTICS_WRITE_KEY, fetch_new_access_token
    import segment.analytics as analytics

    analytics.write_key = ANALYTICS_WRITE_KEY

    user_id = keyring.get_password(SERVICE_ID, 'user_id')
    expires_at = keyring.get_password(SERVICE_ID, 'expires_at')
    if not user_id and ctx.invoked_subcommand != "login":
        typer.echo("Please login first. Run `mirageml login`")
        raise typer.Exit()
    elif expires_at and float(expires_at) < time.time() and ctx.invoked_subcommand != "login":
        try:
            fetch_new_access_token()
            analytics.identify(user_id)
        except:
            typer.echo("Please login again. Run `mirageml login`")
            raise typer.Exit()

    full_command = " ".join(sys.argv[1:])
    analytics.track(user_id, "command", {"command": full_command})


@app.command(name="help", hidden=True)
def custom_help():
    # Your custom help message
    import os
    os.system("mml --help")


@app.command(name="login")
def login_command():
    """Login to Mirage ML"""
    from .commands import login
    login()


def generate_help_text():
    from .constants import help_list_sources
    return help_list_sources()

@app.command()
def chat(
        file_or_url: str = typer.Option(None, "--file-or-url", "-f", help="Path to a file or URL to use as context. \n\n\n**"+sys.argv[0].split('/')[-1]+" chat -f {filepath_or_url}**"),
        sources: List[str] = typer.Option([], "--sources", "-s", help=generate_help_text())
    ):
    """Chat with MirageML"""
    typer.secho("Starting chat. Type 'exit' to end the chat.", fg=typer.colors.BRIGHT_GREEN, bold=True)
    if sources:
        from .commands import rag_chat
        rag_chat(file_or_url, sources)
    else:
        from .commands import normal_chat
        normal_chat(file_or_url=file_or_url)


@config_app.command(name="show")
def show_config_command():
    """Show the current config"""
    from .commands import show_config
    show_config()


@config_app.command(name="set")
def set_config_command():
    """
    Set the config file for MirageML
    """
    from .commands import set_config
    set_config()


# List Commands
@list_app.command(name="plugins")
def list_plugins_command():
    """List connected plugins"""
    from .commands import list_plugins
    list_plugins()


@list_app.command(name="sources")
def list_sources_command():
    """List created sources"""
    from .commands import list_sources
    list_sources()


# Add Commands
@add_app.command(name="plugin")
def add_plugin_command(name: str):
    """Add a plugin by name"""
    from .commands import add_plugin
    add_plugin({"plugin": name})


@add_app.command(name="source")
def add_source_command():
    """Add a new source"""
    from .commands import add_source
    while True:
        link = input("Link for the source: ")
        if not link.startswith("https://"):
            typer.echo("Please enter a valid link starting with https://")
            continue
        break
    from urllib.parse import urlparse
    parsed_url = urlparse(link)
    name = parsed_url.netloc.split('.')[0]
    if name == "docs": name = parsed_url.netloc.split('.')[1]
    name = input(f"Name for the source [default: {name}]: ") or name
    add_source(name, link)


# Delete Commands
@delete_app.command(name="source")
def delete_source_command(name: str = typer.Argument(default="", help="Name of the source to delete")):
    """Delete a source"""
    from .commands import delete_source
    delete_source(name)


# Sync Commands
@sync_app.command(name="plugin")
def sync_plugin_command(name: str):
    """Sync a plugin"""
    from .commands import sync_plugin
    sync_plugin({"plugin": name})


if __name__ == "__main__":
    app()



import http.server
import threading
import platform
import subprocess
import sys
import keyring
import jwt
import urllib.parse
import requests
import typer
import segment.analytics as analytics
from mirageml.constants import SERVICE_ID, PORT, ANALYTICS_WRITE_KEY, supabase, SUPABASE_KEY, SUPABASE_URL, NOTION_SYNC_ENDPOINT

analytics.write_key = ANALYTICS_WRITE_KEY
class LoginManager:
  def __init__(self, handler="mirage_auth_handler", provider="google", provider_options={}):
    self._server = None
    self._thread = None
    self._handler = handler
    self._provider = provider
    self._provider_options = provider_options

  def start_web_server(self):
    """Kick off a thread for the local webserver."""
    th = threading.Thread(target=self._start_local_server)
    th.start()
    self._thread = th

  def select_handler(self):
    if self._handler == "mirage_auth_handler":
      return Handler
    elif self._handler == "google_auth_handler":
      return GoogleHandler
    elif self._handler == "notion_auth_handler":
      return NotionHandler

  def _start_local_server(self):
    self._server = http.server.HTTPServer(('localhost', PORT), self.select_handler())
    self._server.serve_forever()

  def open_browser(self):
    """Opens the browser to the login page."""
    # Waits for the server to start.
    while self._server is None:
        oauth_response = supabase.auth.sign_in_with_oauth({
           "provider": self._provider,
           "options": self._provider_options,
        })
        url = oauth_response.url
        system = platform.system()
        if system == 'Darwin':
            cmd = ['open', url]
        elif system == 'Linux':
            cmd = ['xdg-open', url]
        elif system == 'Windows':
            cmd = ['cmd', '/c', 'start', url.replace('&', '^&')]
        else:
            raise RuntimeError(f'Unsupported system: {system}')
        subprocess.run(cmd, check=True)

class Handler(http.server.SimpleHTTPRequestHandler):
  """Handle the response from accounts.google.com."""

  # Sketchy static variable to hold response.
  info = None

  def log_message(self, format, *args):
    return
    # return super().log_message(format, *args)

  def update_keyring(self, info):
    access_token = info['access_token']
    decoded = jwt.decode(access_token, algorithms=["HS256"], options={ "verify_signature": False })
    keyring.set_password(SERVICE_ID, 'access_token', info['access_token'])
    keyring.set_password(SERVICE_ID, 'refresh_token', info['refresh_token'])
    keyring.set_password(SERVICE_ID, 'expires_at', info['expires_at'])
    keyring.set_password(SERVICE_ID, 'user_id', decoded["sub"])
    keyring.set_password(SERVICE_ID, 'email', decoded["email"])
    analytics.identify(decoded["sub"], {
      "email": decoded["email"],
    })
    return access_token, decoded["sub"]

  def callback_handler(self):
    self.send_response(200)
    self.send_header('Content-type', 'text/html')
    self.end_headers()
    # Serve an HTML page with JavaScript to send the fragment to the server
    self.wfile.write(b"""
    <html>
        <body>
            <script>
                // Send the fragment to the server
                fetch('/capture_fragment?' + location.hash.substr(1))
                .then(() => {
                    document.body.innerHTML = 'All set, feel free to close this tab';
                    window.close();
                });
            </script>
        </body>
    </html>
    """)

  def capture_fragment_handler(self):
    self.send_response(200)
    self.send_header('Content-type', 'text/html')
    self.end_headers()
    self.wfile.write(b'All set, feel free to close this tab')
    # parse the fragment for the access token and refresh token
    fragment = self.path.split('?')[1]
    info = dict(kv.split('=') for kv in fragment.split('&'))
    self.update_keyring(info)
    sys.exit(0)

  def do_GET(self):
    if self.path.startswith("/callback"):
      self.callback_handler()
    elif self.path.startswith("/capture_fragment"):
      self.capture_fragment_handler()

class GoogleHandler(Handler):

  def capture_fragment_handler(self):
    self.send_response(200)
    self.send_header('Content-type', 'text/html')
    self.end_headers()
    self.wfile.write(b'All set, feel free to close this tab')
    # parse the fragment for the access token and refresh token
    fragment = self.path.split('?')[1]
    info = dict(kv.split('=') for kv in fragment.split('&'))
    access_token, user_id = self.update_keyring(info)

    provider_token = info['provider_token']
    provider_refresh_token = urllib.parse.unquote(info['provider_refresh_token'])

    if provider_token and provider_refresh_token:
      params = {
        "user_id": user_id,
        "provider_token": provider_token,
        "provider_refresh_token": provider_refresh_token,
      }
      headers = {
        "apikey": SUPABASE_KEY,
        "Authorization": f"Bearer {access_token}",
        "Content-Type": "application/json",
        "Prefer": "resolution=merge-duplicates",
      }
      requests.post(f"{SUPABASE_URL}/rest/v1/user_google_tokens", json=params, headers=headers)

    sys.exit(0)

class NotionHandler(Handler):
  def capture_fragment_handler(self):
    self.send_response(200)
    self.send_header('Content-type', 'text/html')
    self.end_headers()
    self.wfile.write(b'All set, feel free to close this tab')
    # parse the fragment for the access token and refresh token
    fragment = self.path.split('?')[1]
    info = dict(kv.split('=') for kv in fragment.split('&'))
    access_token, user_id = self.update_keyring(info)
    provider_token = info['provider_token']
    if provider_token:
      keyring.set_password(SERVICE_ID, 'notion_provider_token', info['provider_token'])
      params = {
        "user_id": user_id,
        "provider_token": provider_token,
      }
      headers = {
        "apikey": SUPABASE_KEY,
        "Authorization": f"Bearer {access_token}",
        "Content-Type": "application/json",
        "Prefer": "resolution=merge-duplicates",
      }
      requests.post(f"{SUPABASE_URL}/rest/v1/user_notion_tokens", json=params, headers=headers)
      headers = {
        "Authorization": f"Bearer {access_token}",
      }
      sync_response = requests.post(NOTION_SYNC_ENDPOINT, json={}, headers=headers)
      sync_response_data = sync_response.json()
      if "error" in sync_response_data:
          typer.secho(f"Error syncing notion: {sync_response_data['error']}", fg=typer.colors.BRIGHT_RED, bold=True)
      else:
          typer.secho("Syncing notion triggered successfully. You will receive an email when the sync is complete.", fg=typer.colors.BRIGHT_GREEN, bold=True)
    sys.exit(0)

  def do_GET(self):
    if self.path.startswith("/callback"):
      self.callback_handler()
    elif self.path.startswith("/capture_fragment"):
      self.capture_fragment_handler()


from setuptools import setup, find_packages

setup(
    name='mirageml',
    version='0.1.0',
    author='Mirage ML Inc',
    author_email='support@mirageml.com',
    description='Your Personal AGI',
    packages=find_packages(),
    install_requires=[
        # List your package's dependencies here
        "rich==13.5.2",
        "typer==0.9.0",
        "chardet==5.2.0",
        "qdrant-client==1.6.0",
        "langchain==0.0.314",
        "playwright==1.38.0",
        "beautifulsoup4==4.12.2",
        "transformers==4.34.0",
        "ctransformers==0.2.27",
        "sentence_transformers==2.2.2",
        "torch==2.0.1",
        "pyjwt==2.8.0",
        "keyring==24.2.0",
        "requests==2.31.0",
        "supabase==1.2.0",
        "segment-analytics-python==2.2.3",
        "pyperclip==1.8.2",
        "prompt-toolkit==3.0.39",
    ],
    entry_points={
        'console_scripts': [
            'mirageml=mirageml.__main__:app',
            'mirage=mirageml.__main__:app',
            'mml=mirageml.__main__:app'        # This is the alias 'mml'
        ]
    }
)



import typer
from rich.box import HORIZONTALS
from rich.markdown import Markdown
from rich.console import Console
from rich.live import Live
from rich.panel import Panel
from rich.prompt import Prompt

from .config import load_config
from .utils.codeblocks import extract_code_from_markdown, add_indices_to_code_blocks, copy_code_to_clipboard
from .utils.multiline import multiline_input
from .utils.brain import llm_call
from .utils.prompt_templates import CHAT_TEMPLATE

console = Console()

def normal_chat(file_or_url: str = None):
    # Try loading in data from file
    file_or_url_data, ai_response = "", ""
    if file_or_url:
        from .utils.web_source import extract_file_or_url
        source, file_or_url_data = extract_file_or_url(file_or_url)

    config = load_config()
    chat_history = [{"role": "system", "content": "You are a helpful assistant."}]

    while True:
        try:
            code_blocks = extract_code_from_markdown(ai_response)
            if code_blocks:
                selected_indices = multiline_input("Enter the numbers of the code blocks you want to copy, separated by commas (e.g. '1,3,4') or ask a follow-up")
                try:
                    selected_indices = list(map(int, selected_indices.split(',')))
                    copy_code_to_clipboard(code_blocks, selected_indices)
                    ai_response = ""
                    continue
                except ValueError: user_input = selected_indices
            else:
                if len(chat_history) == 1: user_input = multiline_input("Chat with Mirage")
                else: user_input = multiline_input("Ask a follow-up")
            if user_input.lower().strip() == 'exit':
                typer.secho("Ending chat. Goodbye!", fg=typer.colors.BRIGHT_GREEN, bold=True)
                break
        except KeyboardInterrupt:
            typer.secho("Ending chat. Goodbye!", fg=typer.colors.BRIGHT_GREEN, bold=True)
            break

        if file_or_url_data and len(chat_history) == 1:
            user_input = CHAT_TEMPLATE.format(context=file_or_url_data, question=user_input)
        chat_history.append({"role": "user", "content": user_input})

        # Show the typing indicator using Live
        with Live(Panel("Assistant is typing...", title="[bold blue]Assistant[/bold blue]", box=HORIZONTALS, border_style="blue"),
                  console=console, transient=True, auto_refresh=True, refresh_per_second=8) as live:
            response = llm_call(chat_history, model=config["model"], stream=True, local=config["local_mode"])

            ai_response = ""
            if config["local_mode"]:
                for chunk in response:
                    ai_response += chunk
                    live.update(Panel(Markdown(ai_response), title="[bold blue]Assistant[/bold blue]", box=HORIZONTALS, border_style="blue"))
            else:
                for chunk in response.iter_content():
                    if chunk:
                        decoded_chunk = chunk.decode('utf-8')
                        ai_response += decoded_chunk
                        live.update(Panel(Markdown(ai_response), title="[bold blue]Assistant[/bold blue]", box=HORIZONTALS, border_style="blue"))

        # Print the final AI response outside of the Live context so it persists
        chat_history.append({"role": "assistant", "content": ai_response})
        indexed_ai_response = add_indices_to_code_blocks(ai_response)
        console.print(Panel(Markdown(indexed_ai_response), title="[bold blue]Assistant[/bold blue]", box=HORIZONTALS, border_style="blue"))



import keyring
import requests
import typer

from mirageml.constants import SERVICE_ID, SUPABASE_URL, SUPABASE_KEY

plugin_mapping = {
    "google_token": "gdrive",
    "notion_token": "notion",
}

def list_plugins():
    access_token = keyring.get_password(SERVICE_ID, 'access_token')
    user_id = keyring.get_password(SERVICE_ID, 'user_id')
    params = { "input_user_id": user_id }
    headers = {
        "apikey": SUPABASE_KEY,
        "Authorization": f"Bearer {access_token}",
        "Content-Type": "application/json",
    }
    response = requests.post(f"{SUPABASE_URL}/rest/v1/rpc/user_plugins", json=params, headers=headers)
    response_data = response.json()[0]
    connected_plugin_string = "Connected plugins: "
    for key in response_data:
        if (response_data[key]):
            connected_plugin_string += f"{plugin_mapping[key]}, "
    typer.secho(connected_plugin_string[:-2], fg=typer.colors.BRIGHT_GREEN, bold=True)